import { isPresent } from './utilities';
import { Validators } from './validators';

export const VALID = 'VALID';
export const INVALID = 'INVALID';
export const PENDING = 'PENDING';
export const DISABLED = 'DISABLED';

export type ValidatorFn = (c: FormControl) => {[key: string]: any};

function composeValidators(validators: ValidatorFn[]): ValidatorFn {
  return isPresent(validators) ? Validators.compose(validators) : null;
}

function coerceToValidator(validator: ValidatorFn | ValidatorFn[]): ValidatorFn {
  return Array.isArray(validator) ? composeValidators(validator) : validator;
}

export class FormControl {
  public validator: ValidatorFn;
  private _value: any;
  private _status: string;
  private _errors: {[key: string]: any};

  get value(): any { return this._value; }

  get disabled(): boolean {
    return this._status === DISABLED;
  }

  get enabled(): boolean {
    return this._status !== DISABLED;
  }

  get status(): string {
    return this._status;
  }

  get valid(): boolean {
    return this._status === VALID;
  }

  get errors(): {[key: string]: any} {
    return this._errors;
  }

  constructor(formState: any = null, validator: ValidatorFn|ValidatorFn[] = null) {
    this.validator = coerceToValidator(validator);
    this._applyFormState(formState);
    this.updateValueAndValidity();
  }

  public disable() {
    this._status = DISABLED;
    this._errors = null;
  }

  public enable() {
    this._status = VALID;
    this._errors = null;
  }

  public setValue(value: any) {
    this._value = value;
    this.updateValueAndValidity();
  }

  public reset(value: any) {
    this._applyFormState(value);
  }

  public updateValueAndValidity(): void {

    if (this.enabled) {
      this._errors = this._runValidator();
      this._status = this._calculateStatus();
    }
  }

  public setValidators(newValidator: ValidatorFn|ValidatorFn[]): void {
    this.validator = coerceToValidator(newValidator);
  }

  /**
   * Empties out the sync validator list.
   */
  public clearValidators(): void { this.validator = null; }

  private _isBoxedValue(formState: any): boolean {
    return typeof formState === 'object' && formState !== null &&
        Object.keys(formState).length === 2 && 'value' in formState && 'disabled' in formState;
  }

  private _applyFormState(formState: any) {
    if (this._isBoxedValue(formState)) {
      this._value = formState.value;
      formState.disabled ?
        this.disable() : this.enable();
    } else {
      this._value = formState;
    }
  }

  private _calculateStatus() {
    if (this.disabled) {
      return DISABLED;
    }
    if (this._errors) {
      return INVALID;
    }
    return VALID;
  }

  private _runValidator() {
    return this.validator ? this.validator(this) : null;
  }
}

// export abstract class AbstractControl {
//   _value: any;

//   private _status: string;
//   private _errors: {[key: string]: any};

//   private _parent: FormGroup;

//   constructor(public validator: ValidatorFn) {}

//   get value(): any { return this._value; }

//   get parent(): FormGroup { return this._parent; }

//   get status(): string { return this._status; }

//   get valid(): boolean { return this._status === VALID; }

//   get invalid(): boolean { return this._status === INVALID; }

//   get pending(): boolean { return this._status == PENDING; }

//   get disabled(): boolean { return this._status === DISABLED; }

//   setValidators(newValidator: ValidatorFn|ValidatorFn[]): void {
//     this.validator = coerceToValidator(newValidator);
//   }

//   clearValidators(): void { this.validator = null; }

//   markAsPending({onlySelf}: {onlySelf?: boolean} = {}): void {
//     this._status = PENDING;

//     if (this._parent && !onlySelf) {
//       this._parent.markAsPending({onlySelf});
//     }
//   }

//   disable({onlySelf, emitEvent}: {onlySelf?: boolean, emitEvent?: boolean} = {}): void {
//     this._status = DISABLED;
//     this._errors = null;
//     this._forEachChild((control: AbstractControl) => { control.disable({onlySelf: true}); });
//     this._updateValue();

//     if (emitEvent !== false) {
//       this._valueChanges.emit(this._value);
//       this._statusChanges.emit(this._status);
//     }

//     this._updateAncestors(onlySelf);
//     this._onDisabledChange.forEach((changeFn) => changeFn(true));
//   }

//   enable({onlySelf, emitEvent}: {onlySelf?: boolean, emitEvent?: boolean} = {}): void {
//     this._status = VALID;
//     this._forEachChild((control: AbstractControl) => { control.enable({onlySelf: true}); });
//     this.updateValueAndValidity({onlySelf: true, emitEvent});

//     this._updateAncestors(onlySelf);
//     this._onDisabledChange.forEach((changeFn) => changeFn(false));
//   }

//   setParent(parent: FormGroup): void { this._parent = parent; }

//   abstract setValue(value: any, options?: Object): void;

//   abstract patchValue(value: any, options?: Object): void;

//   abstract reset(value?: any, options?: Object): void;

//   updateValueAndValidity({onlySelf, emitEvent}: {onlySelf?: boolean, emitEvent?: boolean} = {}):
//       void {
//     this._setInitialStatus();
//     this._updateValue();

//     if (this.enabled) {
//       this._errors = this._runValidator();
//       this._status = this._calculateStatus();

//       if (this._status === VALID || this._status === PENDING) {
//         this._runAsyncValidator(emitEvent);
//       }
//     }

//     if (emitEvent !== false) {
//       this._valueChanges.emit(this._value);
//       this._statusChanges.emit(this._status);
//     }

//     if (this._parent && !onlySelf) {
//       this._parent.updateValueAndValidity({onlySelf, emitEvent});
//     }
//   }

//   _updateTreeValidity({emitEvent}: {emitEvent?: boolean} = {emitEvent: true}) {
//     this._forEachChild((ctrl: AbstractControl) => ctrl._updateTreeValidity({emitEvent}));
//     this.updateValueAndValidity({onlySelf: true, emitEvent});
//   }

//   private _setInitialStatus() { this._status = this._allControlsDisabled() ? DISABLED : VALID; }

//   private _runValidator(): {[key: string]: any} {
//     return this.validator ? this.validator(this) : null;
//   }

// }
