import { FormControl, ValidatorFn } from './core';
import { isPresent } from './utilities';

const EMAIL_REGEXP =
  // tslint:disable-next-line:max-line-length
  /^(?=.{1,254}$)(?=.{1,64}@)[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+(\.[-!#$%&'*+/0-9=?A-Z^_`a-z{|}~]+)*@[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?(\.[A-Za-z0-9]([A-Za-z0-9-]{0,61}[A-Za-z0-9])?)*$/;

function isEmptyInputValue(value: any): boolean {
  // we don't check for string here so it also works with arrays
  return value == null || value.length === 0;
}

export class Validators {

  /**
   * Validator that requires controls to have a non-empty value.
   */
  public static required(control: FormControl): {[key: string]: boolean} {
    return isEmptyInputValue(control.value) ? {'required': true} : null;
  }

  /**
   * Validator that requires control value to be true.
   */
  public static requiredTrue(control: FormControl): {[key: string]: boolean} {
    return control.value === true ? null : {'required': true};
  }

  /**
   * Validator that performs email validation.
   */
  public static email(control: FormControl): {[key: string]: boolean} {
    return EMAIL_REGEXP.test(control.value) ? null : {'email': true};
  }

  public static maxLength(maxLength: number): ValidatorFn {
    return (control: FormControl): {[key: string]: any} => {
      const length: number = control.value ? control.value.length : 0;
      return length > maxLength ?
          {'maxlength': {'requiredLength': maxLength, 'actualLength': length}} :
          null;
    };
  }

  public static minLength(minLength: number): ValidatorFn {
    return (control: FormControl): {[key: string]: any} => {
      if (isEmptyInputValue(control.value)) {
        return null;  // don't validate empty values to allow optional controls
      }
      const length: number = control.value ? control.value.length : 0;
      return length < minLength ?
          {'minlength': {'requiredLength': minLength, 'actualLength': length}} :
          null;
    };
  }



  /**
   * Validator that requires a control to match a regex to its value.
   */
  public static pattern(pattern: string|RegExp): ValidatorFn {
    if (!pattern) {
      return Validators.nullValidator;
    }
    let regex: RegExp;
    let regexStr: string;
    if (typeof pattern === 'string') {
      regexStr = `^${pattern}$`;
      regex = new RegExp(regexStr);
    } else {
      regexStr = pattern.toString();
      regex = pattern;
    }
    return (control: FormControl): {[key: string]: any} => {
      if (isEmptyInputValue(control.value)) {
        return null;  // don't validate empty values to allow optional controls
      }
      const value: string = control.value;
      return regex.test(value) ? null :
        {'pattern': {'requiredPattern': regexStr, 'actualValue': value}};
    };
  }

  /**
   * No-op validator.
   */
  public static nullValidator(c: FormControl): {[key: string]: boolean} { return null; }

  /**
   * Compose multiple validators into a single function that returns the union
   * of the individual error maps.
   */
  public static compose(validators: ValidatorFn[]): ValidatorFn {
    if (!validators) {
      return null;
    }
    const presentValidators = validators.filter(isPresent);
    if (presentValidators.length === 0) {
      return null;
    }

    return (control: FormControl) => {
      return _mergeErrors(_executeValidators(control, presentValidators));
    };
  }
}

/**
 * Wraps Javascript Objects
 */
export class StringMapWrapper {
  public static merge<V>(m1: {[key: string]: V}, m2: {[key: string]: V}): {[key: string]: V} {
    const m: {[key: string]: V} = {};

    for (const k of Object.keys(m1)) {
      m[k] = m1[k];
    }

    for (const k of Object.keys(m2)) {
      m[k] = m2[k];
    }

    return m;
  }
}

function _executeValidators(control: FormControl, validators: ValidatorFn[]): any[] {
  return validators.map(v => v(control));
}

function _mergeErrors(arrayOfErrors: any[]): {[key: string]: any} {
  const res: {[key: string]: any} =
      arrayOfErrors.reduce((res: {[key: string]: any}, errors: {[key: string]: any}) => {
        return isPresent(errors) ? StringMapWrapper.merge(res, errors) : res;
      }, {});
  return Object.keys(res).length === 0 ? null : res;
}
