import test from 'ava';
import { FormControl } from '../src/core';
import { Validators } from '../src/validators';

test('it should default the value to null', t => {
  const c = new FormControl();
  t.is(c.value, null);
});

test('It should support valid boxed values on creation', t => {
  const c = new FormControl({value: 'some val', disabled: true});
  t.true(c.disabled);
  t.is(c.value, 'some val');
  t.is(c.status, 'DISABLED');
});

test('It should set new value', t => {
  const c = new FormControl();
  c.setValue('some val');
  t.is(c.value, 'some val');
});

test('It should reset to a specific value if passed', t => {
  const c = new FormControl('some val');
  t.is(c.value, 'some val');
  c.reset('initial value');
  t.is(c.value, 'initial value');
});

test('It should reset to a specific value if passed with boxed value', t => {
  const c = new FormControl('some val');
  t.is(c.value, 'some val');
  c.reset({value: 'initial value', disabled: false});
  t.is(c.value, 'initial value');
});

test('It should run validator with the initial value', t => {
  const c = new FormControl('value', Validators.required);
  t.true(c.valid);
});

test('It should rerun the validator when the value changes', t => {
  const c = new FormControl('value', Validators.required);
  c.setValue(null);
  t.false(c.valid);
});

test('It should support arrays of validator functions if passed', t => {
  const c = new FormControl('', [Validators.required, Validators.maxLength(3)]);
  t.false(c.valid);
  c.setValue('aaa');
  t.true(c.valid);
});

test('It should return errors', t => {
  const c = new FormControl(null, Validators.required);
  t.deepEqual(c.errors, {'required': true});
});

test('It should set single validator', t => {
  const c = new FormControl(null);
  t.true(c.valid);

  c.setValidators(Validators.required);

  c.setValue(null);
  t.false(c.valid);

  c.setValue('abc');
  t.true(c.valid);
});

test('It should set multiple validators from array', t => {
  const c = new FormControl('');
  t.true(c.valid);

  c.setValidators([Validators.minLength(5), Validators.required]);

  c.setValue('');
  t.false(c.valid);

  c.setValue('abc');
  t.false(c.valid);

  c.setValue('abcde');
  t.true(c.valid);
});

test('It should clear validators', t => {
  const c = new FormControl('', Validators.required);
  t.false(c.valid);

  c.clearValidators();
  t.is(c.validator, null);

  c.setValue('');
  t.true(c.valid);
});

test('should add after clearing', t => {
  const c = new FormControl('', Validators.required);
  t.false(c.valid);

  c.clearValidators();
  t.is(c.validator, null);

  c.setValidators([Validators.required]);
  t.not(c.validator, null);
});
