import test from 'ava';
import { FormControl, FormGroup } from '../src/core';
import { Validators } from '../src/validators';

test('It should be the reduced value of the child controls', t => {
  const g = new FormGroup({
    'a': new FormControl('111'),
    'b': new FormControl('222')
  });
  t.deepEqual(g.value, {a: '111', b: '222'});
});

test('It should be the reduced value of the child controls', t => {
  const g = new FormGroup({});
  t.deepEqual(g.value, {});
});

test('It should support nested groups', t => {
  const g = new FormGroup({
    'one': new FormControl('111'),
    'nested': new FormGroup({'two': new FormControl('222')})
  });
  t.deepEqual(g.value, {'one': '111', 'nested': {'two': '222'}});

  (<FormControl>(g.get('nested.two'))).setValue('333');

  t.deepEqual(g.value, {'one': '111', 'nested': {'two': '333'}});
});

test('It should update value and validity when control is added', t => {
  const g = new FormGroup({'one': new FormControl('1')});
  t.deepEqual(g.value, {'one': '1'});
  t.true(g.value);

  g.addControl('two', new FormControl('2', Validators.minLength(10)));

  t.deepEqual(g.value, {'one': '1', 'two': '2'});
  t.false(g.valid);
});

test('It should update value and validity when control is removed', t => {
  const g = new FormGroup(
    {'one': new FormControl('1'), 'two': new FormControl('2', Validators.minLength(10))});

  t.deepEqual(g.value, {'one': '1', 'two': '2'});
  t.false(g.valid);

  g.removeControl('two');

  t.deepEqual(g.value, {'one': '1', 'two': '2'});
  t.false(g.valid);
});

test('It should run the validator when the value changes', t => {
  const simpleValidator = (c: FormGroup) =>
    c.controls['one'].value !== 'correct' ? {'broken': true} : null;

  const c = new FormControl(null);
  const g = new FormGroup({'one': c}, simpleValidator);

  c.setValue('correct');

  t.true(g.value);
  t.is(g.errors, null);

  c.setValue('incorrect');

  t.false(g.value);
  t.deepEqual(g.errors, {'broken': true});
});

(() => {
  let c: FormControl;
  let c2: FormControl;
  let g: FormGroup;

  test.beforeEach(() => {
    c = new FormControl('');
    c2 = new FormControl('');
    g = new FormGroup({'one': c, 'two': c2});
  });

  test('It should set its own value', t => {
    g.setValue({'one': 'one', 'two': 'two'});
    t.deepEqual(g.value, {'one': 'one', 'two': 'two'});
  });

  test('It should set child values', t => {
    g.setValue({'one': 'one', 'two': 'two'});
    t.is(c.value, 'one');
    t.is(c2.value, 'two');
  });

  test('It should set child control values if disabled', t => {
    c2.disable();
    g.setValue({'one': 'one', 'two': 'two'});

    t.is(c2.value, 'two');
    t.deepEqual(g.value, {'one': 'one'});
    t.deepEqual(g.getRawValue(), {'one': 'one', 'two': 'two'});
  });

  test('It should set group value if group is disabled', t => {
    g.disable();
    g.setValue({'one': 'one', 'two': 'two'});

    t.is(c.value, 'one');
    t.is(c2.value, 'two');

    t.deepEqual(g.value, {'one': 'one', 'two': 'two'});
  });

  test('It should set parent values', t => {
    const form = new FormGroup({'parent': g});
    g.setValue({'one': 'one', 'two': 'two'});
    t.deepEqual(form.value, {'parent': {'one': 'one', 'two': 'two'}});
  });

  test('It should not update the parent when explicitly specified', t => {
    const form = new FormGroup({'parent': g});
    g.setValue({'one': 'one', 'two': 'two'}, {onlySelf: true});
    t.deepEqual(form.value, {parent: {'one': '', 'two': ''}});
  });

  test('It should throw if fields are missing from supplied value (subset)', t => {
    const error = t.throws(() => g.setValue({
      'one': 'one'
    }));
    t.is(error, new RegExp(`Must supply a value for form control with name: 'two'`));
  });

  test('It should throw if a value is provided for a missing control (superset)', t => {
    const error = t.throws(() => g.setValue({'one': 'one', 'two': 'two', 'three': 'three'}));
    t.is(error, new RegExp(`Cannot find form control with name: three`));
  });

  test('It should throw if a value is not provided for a disabled control', t => {
    c2.disable();
    const error = t.throws(() => g.setValue({
      'one': 'one'
    }));
    t.is(error, new RegExp(`Must supply a value for form control with name: 'two'`));
  });

  test('It should throw if no controls are set yet', t => {
    const empty = new FormGroup({});
    const error = t.throws(() => empty.setValue({
      'one': 'one'
    }));
    t.is(error, new RegExp(`no form controls registered with this group`));
  });

})();
