import test from 'ava';
import { FormControl } from '../src/core';
import { Validators } from '../src/validators';

function validator(key: string, error: any) {
  return function(c: FormControl) {
    const r: {[k: string]: string} = {};
    r[key] = error;
    return r;
  };
}

test('required: it should error on an empty string', t => {
  t.deepEqual(Validators.required(new FormControl('')), {'required': true});
});

test('required: it should error on null', t => {
  t.deepEqual(Validators.required(new FormControl(null)), {'required': true});
});

test('required: it should not error on a non-empty string', t => {
  t.is(Validators.required(new FormControl('not empty')), null);
});

test('required: it should accept zero as valid', t => {
  t.is(Validators.required(new FormControl(0)), null);
});

test('required: it should error on an empty array', t => {
  t.deepEqual(Validators.required(new FormControl([])), {'required': true});
});

test('required: it should not error on a non-empty array', t => {
  t.deepEqual(Validators.required(new FormControl([1, 2])), null);
});

test('requiredTrue: It should error on false', t => {
  t.deepEqual(Validators.requiredTrue(new FormControl(false)), {'required': true});
});

test('requiredTrue: It should not error on true', t => {
  t.is(Validators.requiredTrue(new FormControl(true)), null);
});

test('minLength: It should not error on an empty string', t => {
  t.is(Validators.minLength(2)(new FormControl('')), null);
});

test('minLength: It should not error on null', t => {
  t.is(Validators.minLength(2)(new FormControl(null)), null);
});

test('minLength: It should not error on undefined', t => {
  t.is(Validators.minLength(2)(new FormControl(undefined)), null);
});

test('minLength: It should not error on valid strings', t => {
  t.is(Validators.minLength(2)(new FormControl('aa')), null);
});

test('minLength: It should error on short strings', t => {
  t.deepEqual(Validators.minLength(2)(new FormControl('a')), {
    'minlength': {'requiredLength': 2, 'actualLength': 1}
  });
});

test('maxLength: It should not error on an empty string', t => {
  t.is(Validators.maxLength(2)(new FormControl('')), null);
});

test('maxLength: It should not error on null', t => {
  t.is(Validators.maxLength(2)(new FormControl('aa')), null);
});

test('maxLength: It should not error on valid strings', t => {
  t.is(Validators.maxLength(2)(new FormControl('aa')), null);
});

test('maxLength: It should error on long strings', t => {
  t.deepEqual(Validators.maxLength(2)(new FormControl('aaa')), {
    'maxlength': {'requiredLength': 2, 'actualLength': 3}
  });
});

test('pattern: It should not error on an empty string', t => {
  t.is(Validators.pattern('[a-zA-Z ]+')(new FormControl('')), null);
});

test('pattern: It should not error on null', t => {
  t.is(Validators.pattern('[a-zA-Z]+')(new FormControl(null)), null);
});

test('pattern: It should not error on undefined', t => {
  t.is(Validators.pattern('[a-zA-Z]+')(new FormControl()), null);
});

test('pattern: It should not error on null value and "null" pattern', t => {
  t.is(Validators.pattern('null')(new FormControl(null)), null);
});

test('pattern: It should not error on valid strings', t => {
  t.is(Validators.pattern('[a-zA-Z]*')(new FormControl('aaAA')), null);
});

test('pattern: It should error on failure to match string', t => {
  t.deepEqual(Validators.pattern('[a-zA-Z]*')(new FormControl('aaa0')), {
    'pattern': {'requiredPattern': '^[a-zA-Z]*$', 'actualValue': 'aaa0'}
  });
});

test('pattern: It should accept RegExp object', t => {
  const pattern: RegExp = new RegExp('[a-zA-Z]+');
  t.is(Validators.pattern(pattern)(new FormControl('aaAA')), null);
});

test('pattern: It should error on failure to match RegExp object', t => {
  const pattern: RegExp = new RegExp('^[a-zA-Z]*$');
  t.deepEqual(Validators.pattern(pattern)(new FormControl('aaa0')), {
    'pattern': {'requiredPattern': '/^[a-zA-Z]*$/', 'actualValue': 'aaa0'}
  });
});

test('pattern: It should not error on "null" pattern', t => {
  t.is(Validators.pattern(null)(new FormControl('aaAA')), null);
});

test('pattern: It should not error on "undefined" pattern', t => {
  t.is(Validators.pattern(undefined)(new FormControl('aaAA')), null);
});

test('compose: It should return null when given null', t => {
  t.is(Validators.compose(null), null);
});

test('compose: It should collect errors from all the validators', t => {
  const c = Validators.compose([validator('a', true), validator('b', true)]);
  t.deepEqual(c(new FormControl('')), {
    'a': true, 'b': true
  });
});

test('compose: It should run validators left to right', t => {
  const c = Validators.compose([validator('a', 1), validator('a', 2)]);
  t.deepEqual(c(new FormControl('')), {'a': 2});
});

test('compose: It should return null when no errors', t => {
  const c = Validators.compose([Validators.nullValidator, Validators.nullValidator]);
  t.is(c(new FormControl('')), null);
});

test('compose: It should ignore nulls', t => {
  const c = Validators.compose([null, Validators.required]);
  t.deepEqual(c(new FormControl('')), {'required': true});
});
